#!/usr/bin/perl
use 5.010;
use strict;
use warnings;
use Time::HiRes;


# Globals
#our $ROMSTART = 060000;
#our $ROMEND   = 077777; 

our %sym_tab;			# Symbol Table
our $progname;			# base program name (no extension)
#our $hexlofh;			# Low byte hex file handle
#our $hexhifh;			# High byte hex file handle
our $binfh;			# Absolute Binary File handle
our %devs = (			# I/O Device Codes (Reserved keywords)
"TTI"       => 010,		# Teletype input
"TTO"       => 011,		# Teletype output
"PTR"       => 012,		# Paper tape reader
"PTP"       => 013,		# Paper tape punch
"RTC"       => 014,		# Real time clock
"PLT"       => 015,		# Incremental plotter
"CDR"       => 016,		# Card reader
"LPT"       => 017,		# Line printer
"DSK"       => 020,		# Fixed head disk
"DKP"       => 033,		# Moving head disk
"CAS"       => 034,		# Cassette tape
"DSK1"      => 060,		# Second fixed head disk
"DKP1"      => 073,		# Second moving head disk
"FPU1"      => 074,		# \
"FPU2"      => 075,		#  > Floating point option
"FPU"       => 076,		# /
"CPU"       => 077,		# Central processor and console functions
);

our $memaddr;			# Memory address of current data block
our $wcount;			# Number of words in current data block
our @datablk;			# Array to hold current data block

print "NOVA 445 Absolute Binary Assembler V0.00\n";

if ($#ARGV<0) {
	print "Usage: $0 <src_file>\n";
	exit 1;
}

$progname = $ARGV[0];if ($progname =~ '.') {
	$progname = substr($ARGV[0],0,rindex($ARGV[0],'.'));
}

pass1();
pass2();

print "$ARGV[0] assembled successfully\n";

exit(0);

#######################################################################################
# Pass 1: Build Symbol Table
#######################################################################################
sub pass1 {
	print "==Pass 1==\n";
	my $start = Time::HiRes::gettimeofday();
	my $num_statements = 0;
	my $addr = 0;
	my $linenum = 0;
	my $line;
	
	# Open the input file
	open(my $in, "<", $ARGV[0]) or die "Can't open $ARGV[0]: $!";
	
	while (<$in>) {
		$linenum++;
		# Strip off comments & leading/trailing white space
		my $comment_start = index($_,";");
		
		if ($comment_start < 0) {
			$comment_start = length;
		}
		$line = trim(substr($_,0,$comment_start));
		# Ignore blank lines
		next if $line eq "";

		# Process labels
		my $cend = index($line,':');
		if ($cend > 0) {
			# This line contains a label
			my $label = substr($line,0,$cend);
			# Save label and its address in symbol table
			# First check label for validity
			die "Line $linenum: Invalid label: $label\n" unless $label =~ /^[\.A-Z][\.A-Z0-9]*$/;

			# Check for reserved symbol
			die "Line $linenum: Reserved symbol: $label\n" if exists $main::devs{$label};
			# FixFixFix Shouldn't allow any instruction symbols either
			# Now check for duplicate label
			die "Line $linenum: Duplicate label: $label\n" if exists $main::sym_tab{$label};
			$main::sym_tab{$label} = $addr;

			# Strip the label and process the rest of the line
			$line = trim(substr($line,$cend+1));
			next if $line eq "";
		} elsif ($cend == 0) {
			die "Line $linenum: Syntax error:\n$_\n";
		}

		# Process directives
		if (substr($line,0,4) eq ".END") {
			$line = trim(substr($line,4));
			if ($line ne "") {
				# FixFixFix  This needs to be supported
				printf("Line $linenum: Start addr not supported in .END\n");
				die;
			}
			last;
		}
		if (substr($line,0,4) eq ".LOC") {
			$line = trim(substr($line,4));
			die "Line $linenum: .LOC must be octal" unless $line =~ /^[0-7]+$/;
			$addr = oct($line);
			
			unless (($addr >= 0) && ($addr <= 0177777)) {
				printf("Line $linenum: Invalid address specified: %X\n",$addr);
				die;
			}
			next;
		}
		if (substr($line,0,4) eq ".BLK") {
			$line = trim(substr($line,4));
			if ($line =~ /^\d+\.$/) {
				$line =~ s/\.//;
				$addr += $line;
			} else {
				$addr += oct $line;
			}
			next;
		}
#		die "Line $linenum: Data only allowed in ROM space" unless ($addr >= $ROMSTART) && ($addr <= $ROMEND);
		if (substr($line,0,4) eq ".TXT") {
			# Determine size of string & adjust $addr
			$line = trim(substr($line,4));
			# String should begin and end with "
			die "Line $linenum: missing \" in .TXT string\n" unless $line =~ /^\".*\"$/;
			$line = substr($line,1,-1);
			$line = debracket($line);
			$addr += int((length($line) + 2)/2);
			next;
		}
		if (substr($line,0,5) eq ".INFO") {
			# Ignore .INFO lines on first pass
			next;
		}
		
		
		# Process statements.  Assume that we have a statement at this point -
		# we'll catch syntax errors in pass 2... 
#		die "Line $linenum: Program exceeds ROM size\n" if $addr > $ROMEND;
		$addr++;
		$num_statements++;
	}
	

	my $numsym = keys %main::sym_tab;
	my $end = Time::HiRes::gettimeofday();
	printf "Elapsed time: %.3f\n", $end-$start;
	print "Program contains $numsym symbol(s), $num_statements statement(s).\n\n";
	close $in or die "$in: $!";
}


#######################################################################################
# Pass 2: Generate binary file and listing file
#######################################################################################
sub pass2 {
	my $start   = Time::HiRes::gettimeofday();
	my $addr    = 0;
	my $linenum = 0;
	my $line;
	my $token;
	my $cend;
	my $instr;
	my $isbranch;
	my $isor;
	my $regs;
	my $offset;
	my $i;
	my $c;
	my $first;
	my $indirect;
	my $noload;
	my %instr0 = (			# Instruction that require no arguments
	"MUL"    => 073301,		# Unsigned Multiply
	"MULS"   => 077201,		# Signed Multiply
	"DIV"    => 073101,		# Unsigned Divide
	"DIVS"   => 077001,		# Signed Divide
	"NORM"   => 063301,		# Normalize
	"SLLD"   => 063001,		# Shift Logically Left
	"SALD"   => 073001,		# Shift Arithmetically Left
	"SLRD"   => 063201,		# Shift Logically Right
	"SARD"   => 063101,		# Shift Arithmetically Right
	"SKNV"   => 067301,		# Skip on Not Overflow
	"PSHF"   => 072501,		# Push Flags
	"POPF"   => 072401,		# Pop Flags
	"POPJ"   => 072601,		# Pop PC and Jump
	"PSHR"   => 072701,		# Push Return Address
	"SAV"    => 062401,		# Save
	"RET"    => 062601,		# Return
	"DSP"    => 073201,		# Decrement Stack Pointer
	"INTEN"  => 060177,		# Interrupt Enable
	"INTDS"  => 060277,		# Interrupt Disable
	"IORST"  => 062677,		# I/O Reset
	"HALT"   => 063077,		# Halt the Processor
	"WAIT"   => 067001,		# Wait for Interrupt
	"TRAP"   => 0100010,		# Trap
	"ETRP"   => 077101,		# Enable Trap Instruction
	"DTRP"   => 077301,		# Disable Trap Instruction
	"E64K"   => 067101,		# Enable 64K-words Mode
	"D64K"   => 067201,		# Disable 64K-words Mode
	);
	my %instrac34 = (		# Single arg: AC in bits 3,4
	"PSHA"   => 061401,		# Push Accumulator
	"POPA"   => 061601,		# Pop Accumulator
	"TOPR"   => 061701,		# Read Top of Stack
	"TOPW"   => 061501,		# Write Top of Stack
	"MTSP"   => 061001,		# Move To Stack Pointer
	"MFSP"   => 061201,		# Move From Stack Pointer
	"MTFP"   => 060001,		# Move To Frame Pointer
	"MFFP"   => 060201,		# Move From Frame Pointer
	"READS"  => 060477,		# Read Console Switch Register
	"INTA"   => 061477,		# Interrupt Acknowledge
	"MSKO"   => 062077,		# Mask Out
	);
	my %instr2ac = (		# AC in bits 3,4 & 8,9
	"LDB"    => 060401,		# Load Byte
	"STB"    => 062001,		# Store Byte
	"OR"     => 063401,		# Logical OR (registers reversed!)
	);
	my %instrdev = (		# Device code in bits 10-15
	"NIO"    => 060000,		# No Data Transfer
	"NIOS"   => 060100,		# No Data Transfer, Start
	"NIOC"   => 060200,		# No Data Transfer, Clear
	"NIOP"   => 060300,		# No Data Transfer, Pulse
	"SKPBN"  => 063400,		# Skip if Busy is Non-Zero
	"SKPBZ"  => 063500,		# Skip if Busy is Zero
	"SKPDN"  => 063600,		# Skip if Done is Non-Zero
	"SKPDZ"  => 063700,		# Skip if Done is Zero
	);
	my %instrio = (			# AC in bits 3,4 + Device in 10-15
	"DIA"    => 060400,		# Skip if negative
	"DIAS"   => 060500,		# Skip if positive
	"DIAC"   => 060600,
	"DIAP"   => 060700,
	"DOA"    => 061000,
	"DOAS"   => 061100,
	"DOAC"   => 061200,
	"DOAP"   => 061300,
	"DIB"    => 061400,		# Skip if negative
	"DIBS"   => 061500,		# Skip if positive
	"DIBC"   => 061600,
	"DIBP"   => 061700,
	"DOB"    => 062000,
	"DOBS"   => 062100,
	"DOBC"   => 062200,
	"DOBP"   => 062300,
	"DIC"    => 062400,		# Skip if negative
	"DICS"   => 062500,		# Skip if positive
	"DICC"   => 062600,
	"DICP"   => 062700,
	"DOC"    => 063000,
	"DOCS"   => 063100,
	"DOCC"   => 063200,
	"DOCP"   => 063300,
	);
	my %instrtgt = (		# Just specify target addr
	"JMP"    => 000000,		# Jump to effective addr
	"JSR"    => 004000,		# Jump to subroutine
	"ISZ"    => 010000,		# Increment & Skip if Zero
	"DSZ"    => 014000,		# Decrement & Skip if Zero
	);
	my %instractgt = (		# AC in bits 3,4 + target addr
	"LDA"    => 020000,		# Load Accumulator
	"STA"    => 040000,		# Store Accumulator
	);
	my %instralu = (		# Arithmetic & Logic Insturctions
	"COM"    => 0100000,		# Complement
	"COML"   => 0100100,
	"COMR"   => 0100200,
	"COMS"   => 0100300,
	"COMZ"   => 0100020,
	"COMZL"  => 0100120,
	"COMZR"  => 0100220,
	"COMZS"  => 0100320,
	"COMO"   => 0100040,
	"COMOL"  => 0100140,
	"COMOR"  => 0100240,
	"COMOS"  => 0100340,
	"COMC"   => 0100060,
	"COMCL"  => 0100160,
	"COMCR"  => 0100260,
	"COMCS"  => 0100360,
	"NEG"    => 0100400,		# Negate
	"NEGL"   => 0100500,
	"NEGR"   => 0100600,
	"NEGS"   => 0100700,
	"NEGZ"   => 0100420,
	"NEGZL"  => 0100520,
	"NEGZR"  => 0100620,
	"NEGZS"  => 0100720,
	"NEGO"   => 0100440,
	"NEGOL"  => 0100540,
	"NEGOR"  => 0100640,
	"NEGOS"  => 0100740,
	"NEGC"   => 0100460,
	"NEGCL"  => 0100560,
	"NEGCR"  => 0100660,
	"NEGCS"  => 0100760,
	"MOV"    => 0101000,		# Move
	"MOVL"   => 0101100,
	"MOVR"   => 0101200,
	"MOVS"   => 0101300,
	"MOVZ"   => 0101020,
	"MOVZL"  => 0101120,
	"MOVZR"  => 0101220,
	"MOVZS"  => 0101320,
	"MOVO"   => 0101040,
	"MOVOL"  => 0101140,
	"MOVOR"  => 0101240,
	"MOVOS"  => 0101340,
	"MOVC"   => 0101060,
	"MOVCL"  => 0101160,
	"MOVCR"  => 0101260,
	"MOVCS"  => 0101360,
	"INC"    => 0101400,		# Increment
	"INCL"   => 0101500,
	"INCR"   => 0101600,
	"INCS"   => 0101700,
	"INCZ"   => 0101420,
	"INCZL"  => 0101520,
	"INCZR"  => 0101620,
	"INCZS"  => 0101720,
	"INCO"   => 0101440,
	"INCOL"  => 0101540,
	"INCOR"  => 0101640,
	"INCOS"  => 0101740,
	"INCC"   => 0101460,
	"INCCL"  => 0101560,
	"INCCR"  => 0101660,
	"INCCS"  => 0101760,
	"ADC"    => 0102000,		# Add Complement
	"ADCL"   => 0102100,
	"ADCR"   => 0102200,
	"ADCS"   => 0102300,
	"ADCZ"   => 0102020,
	"ADCZL"  => 0102120,
	"ADCZR"  => 0102220,
	"ADCZS"  => 0102320,
	"ADCO"   => 0102040,
	"ADCOL"  => 0102140,
	"ADCOR"  => 0102240,
	"ADCOS"  => 0102340,
	"ADCC"   => 0102060,
	"ADCCL"  => 0102160,
	"ADDCR"  => 0102260,
	"ADDCS"  => 0102360,
	"SUB"    => 0102400,		# Subtract
	"SUBL"   => 0102500,
	"SUBR"   => 0102600,
	"SUBS"   => 0102700,
	"SUBZ"   => 0102420,
	"SUBZL"  => 0102520,
	"SUBZR"  => 0102620,
	"SUBZS"  => 0102720,
	"SUBO"   => 0102440,
	"SUBOL"  => 0102540,
	"SUBOR"  => 0102640,
	"SUBOS"  => 0102740,
	"SUBC"   => 0102460,
	"SUBCL"  => 0102560,
	"SUBCR"  => 0102660,
	"SUBCS"  => 0102760,
	"ADD"    => 0103000,		# Add
	"ADDL"   => 0103100,
	"ADDR"   => 0103200,
	"ADDS"   => 0103300,
	"ADDZ"   => 0103020,
	"ADDZL"  => 0103120,
	"ADDZR"  => 0103220,
	"ADDZS"  => 0103320,
	"ADDO"   => 0103040,
	"ADDOL"  => 0103140,
	"ADDOR"  => 0103240,
	"ADDOS"  => 0103340,
	"ADDC"   => 0103060,
	"ADDCL"  => 0103160,
	"ADDCR"  => 0103260,
	"ADDCS"  => 0103360,
	"AND"    => 0103400,		# Logical AND
	"ANDL"   => 0103500,
	"ANDR"   => 0103600,
	"ANDS"   => 0103700,
	"ANDZ"   => 0103420,
	"ANDZL"  => 0103520,
	"ANDZR"  => 0103620,
	"ANDZS"  => 0103720,
	"ANDO"   => 0103440,
	"ANDOL"  => 0103540,
	"ANDOR"  => 0103640,
	"ANDOS"  => 0103740,
	"ANDC"   => 0103460,
	"ANDCL"  => 0103560,
	"ANDCR"  => 0103660,
	"ANDCS"  => 0103760,
	);
	my %rskip = (			# Registers + skip
	"0,0"     => 000000,
	"0,0,SKP" => 000001,
	"0,0,SZC" => 000002,
	"0,0,SNC" => 000003,
	"0,0,SZR" => 000004,
	"0,0,SNR" => 000005,
	"0,0,SEZ" => 000006,
	"0,0,SBN" => 000007,
	"0,1"     => 004000,
	"0,1,SKP" => 004001,
	"0,1,SZC" => 004002,
	"0,1,SNC" => 004003,
	"0,1,SZR" => 004004,
	"0,1,SNR" => 004005,
	"0,1,SEZ" => 004006,
	"0,1,SBN" => 004007,
	"0,2"     => 010000,
	"0,2,SKP" => 010001,
	"0,2,SZC" => 010002,
	"0,2,SNC" => 010003,
	"0,2,SZR" => 010004,
	"0,2,SNR" => 010005,
	"0,2,SEZ" => 010006,
	"0,2,SBN" => 010007,
	"0,3"     => 014000,
	"0,3,SKP" => 014001,
	"0,3,SZC" => 014002,
	"0,3,SNC" => 014003,
	"0,3,SZR" => 014004,
	"0,3,SNR" => 014005,
	"0,3,SEZ" => 014006,
	"0,3,SBN" => 014007,
	"1,0"     => 020000,
	"1,0,SKP" => 020001,
	"1,0,SZC" => 020002,
	"1,0,SNC" => 020003,
	"1,0,SZR" => 020004,
	"1,0,SNR" => 020005,
	"1,0,SEZ" => 020006,
	"1,0,SBN" => 020007,
	"1,1"     => 024000,
	"1,1,SKP" => 024001,
	"1,1,SZC" => 024002,
	"1,1,SNC" => 024003,
	"1,1,SZR" => 024004,
	"1,1,SNR" => 024005,
	"1,1,SEZ" => 024006,
	"1,1,SBN" => 024007,
	"1,2"     => 030000,
	"1,2,SKP" => 030001,
	"1,2,SZC" => 030002,
	"1,2,SNC" => 030003,
	"1,2,SZR" => 030004,
	"1,2,SNR" => 030005,
	"1,2,SEZ" => 030006,
	"1,2,SBN" => 030007,
	"1,3"     => 034000,
	"1,3,SKP" => 034001,
	"1,3,SZC" => 034002,
	"1,3,SNC" => 034003,
	"1,3,SZR" => 034004,
	"1,3,SNR" => 034005,
	"1,3,SEZ" => 034006,
	"1,3,SBN" => 034007,
	"2,0"     => 040000,
	"2,0,SKP" => 040001,
	"2,0,SZC" => 040002,
	"2,0,SNC" => 040003,
	"2,0,SZR" => 040004,
	"2,0,SNR" => 040005,
	"2,0,SEZ" => 040006,
	"2,0,SBN" => 040007,
	"2,1"     => 044000,
	"2,1,SKP" => 044001,
	"2,1,SZC" => 044002,
	"2,1,SNC" => 044003,
	"2,1,SZR" => 044004,
	"2,1,SNR" => 044005,
	"2,1,SEZ" => 044006,
	"2,1,SBN" => 044007,
	"2,2"     => 050000,
	"2,2,SKP" => 050001,
	"2,2,SZC" => 050002,
	"2,2,SNC" => 050003,
	"2,2,SZR" => 050004,
	"2,2,SNR" => 050005,
	"2,2,SEZ" => 050006,
	"2,2,SBN" => 050007,
	"2,3"     => 054000,
	"2,3,SKP" => 054001,
	"2,3,SZC" => 054002,
	"2,3,SNC" => 054003,
	"2,3,SZR" => 054004,
	"2,3,SNR" => 054005,
	"2,3,SEZ" => 054006,
	"2,3,SBN" => 054007,
	"3,0"     => 060000,
	"3,0,SKP" => 060001,
	"3,0,SZC" => 060002,
	"3,0,SNC" => 060003,
	"3,0,SZR" => 060004,
	"3,0,SNR" => 060005,
	"3,0,SEZ" => 060006,
	"3,0,SBN" => 060007,
	"3,1"     => 064000,
	"3,1,SKP" => 064001,
	"3,1,SZC" => 064002,
	"3,1,SNC" => 064003,
	"3,1,SZR" => 064004,
	"3,1,SNR" => 064005,
	"3,1,SEZ" => 064006,
	"3,1,SBN" => 064007,
	"3,2"     => 070000,
	"3,2,SKP" => 070001,
	"3,2,SZC" => 070002,
	"3,2,SNC" => 070003,
	"3,2,SZR" => 070004,
	"3,2,SNR" => 070005,
	"3,2,SEZ" => 070006,
	"3,2,SBN" => 070007,
	"3,3"     => 074000,
	"3,3,SKP" => 074001,
	"3,3,SZC" => 074002,
	"3,3,SNC" => 074003,
	"3,3,SZR" => 074004,
	"3,3,SNR" => 074005,
	"3,3,SEZ" => 074006,
	"3,3,SBN" => 074007,
	);
	my %regsd = (			# Source/Destination registers
	"0,0"  => 000000,
	"0,1"  => 004000,
	"0,2"  => 010000,
	"0,3"  => 014000,
	"1,0"  => 000100,
	"1,1"  => 004100,
	"1,2"  => 010100,
	"1,3"  => 014100,
	"2,0"  => 000200,
	"2,1"  => 004200,
	"2,2"  => 010200,
	"2,3"  => 014200,
	"3,0"  => 000300,
	"3,1"  => 004300,
	"3,2"  => 010300,
	"3,3"  => 014300,
	);

	print "==Pass 2==\n";

	$instr = 0;			# DEBUG
	$memaddr = 0;
	# Open the input file
	open(my $in, "<", $ARGV[0]) or die "Can't open $ARGV[0]: $!";
	# Open Listing file and binary file.
	my $listname = $progname . '.LST';
	open(my $listfh, ">", $listname) or die "Can't open listing file: $listname: $!";
#	openhexfiles();
	openbinfile();

	while (<$in>) {
		$linenum++;
		# Strip off comments & leading/trailing white space
		my $comment_start = index($_,";");
		
		if ($comment_start < 0) {
			$comment_start = length;
		}
		$line = trim(substr($_,0,$comment_start));
		# Ignore blank lines
		if ($line eq "") {
			printf {$listfh} "%04d                    %s",$linenum,$_;
			next;
		}

		# Process directives
		if (substr($line,0,5) eq ".INFO") {
			flushbin();
			printf {$listfh} "%04d                    %s",$linenum,$_;
			# Remove excess white space & add trailing rubout byte
			$line = trim(substr($line,5));
			# Write IGNORE block to binary file
			writeword(0x0A0D);			# CRLF indicates IGNORE block
			syswrite $binfh, $line,length($line);	# The text
			writebyte(0xFF);			# RUBOUT ends block
			next;
		}
		if (substr($line,0,4) eq ".END") {
			# FixFixFix Allow starting addr to be specified!
			printf {$listfh} "%04d                    %s",$linenum,$_;
			flushbin();
			writeword(0x0001);			# START block
			writeword(0x8000);			# No start address
			writeword(0x7FFF);			# checksum value
			last;
		}
		if (substr($line,0,4) eq ".LOC") {
			# We already checked that the address was valid on the first pass
			$addr = oct(trim(substr($line,4)));
			flushbin();
			$memaddr = $addr;			# start of new data block
			printf {$listfh} "%04d %06o              %s",$linenum,$addr,$_;
			next;
		}
		# Process labels (by ignoring them on this pass)
		$cend = index($line,':');
		if ($cend > 0) {
			# This line contains a label
			# Strip the label and process the rest of the line
			$line = trim(substr($line,$cend+1));
			if ($line eq "") {
				printf {$listfh} "%04d                 %s",$linenum,$_;
				next;
			}
		}
		if (substr($line,0,4) eq ".TXT") {
			$first = 1;
			# Generate binary for null-terminated text string
			# As this is written, it is not necessary to escape quotes within
			# a string.
			$line = trim(substr($line,4));
			# Already verified that string begins and ends with "
			# Add trailing null byte(s)
			$line = substr($line,1,-1) . chr(0);
			$line = debracket($line);
			if (length($line) & 1) {
				$line = $line . chr(0);
			}
			for ($i=0; $i<length($line); $i+=2) {
				$c = substr($line,$i,1);
				$instr = ord($c);
				$c = substr($line,$i+1,1);
				$instr |= (ord($c) << 8);
				if ($first) {
					printf {$listfh} "%04d %06o %06o      %s",$linenum,$addr,$instr,$_;
					$first = "";
				} else {
					printf {$listfh} "     %06o %06o\n",$addr,$instr;
				}
#				writehex($addr,$instr);
				writebin($addr, $instr);
				$addr++;
			}
			next;
		}
		if (substr($line,0,4) eq ".BLK") {
			$token = trim(substr($line,4));
			printf {$listfh} "%04d %06o             %s",$linenum,$addr,$_;
			if ($token =~ /^\d+\.$/) {
				$token =~ s/\.//;
				$addr += $token;
			} else {
				die "Line $linenum: Invalid .BLK size: $token" unless $token =~ /^[0-7]+$/;
				$addr += oct $token;
			}
			next;
		}

		# Get next token
		# Convert white space chars to space
		$line =~ s/\s+/ /g;
		# Check for indirect '@' (anywhere on line) & remove
		if ($line =~ /@/) {
			$indirect = 002000;
			$line =~ s/@//g;
		} else {
			$indirect = 0;
		}
		# Same for no-load '#'
		if ($line =~ /\#/) {
			$noload = 1;
			$line =~ s/\#//g;
		} else {
			$noload = 0;
		}
 

		$token = $line;
		my $zzz = index($line, ' ');
		if ($zzz > 0) {
			$token = substr($line,0,$zzz);
		}
		# Ignore blank lines
		if ($line eq "") {
			printf {$listfh} "%04d                   %s",$linenum,$_;
			next;
		}

		# Check for numeric value or label - @ sets high bit
		if ($token =~ /^\d/) {
			die "Line: $linenum: No-Load not allowed" unless ($noload eq 0);
			if ($token =~ /^\d+\.$/) {
				$token =~ s/\.//;
				$instr = $token;
			} else {
				die "Line $linenum: Invalid octal value: $token" unless $token =~ /^[0-7]+$/;
				$instr = oct $token;
			}
			die "Line: $linenum: Invalid numeric value" unless ($instr <= 0177777);
			if ($indirect ne 0) {
				$instr |= 0100000;
			}
		} elsif (exists $sym_tab{$token}) {
			die "Line: $linenum: No-Load not allowed" unless ($noload eq 0);
			$instr = $sym_tab{$token};
			if ($indirect ne 0) {
				$instr |= 0100000;
			}
			# Did we gobble the whole line?
			die "Line $linenum: Syntax error:\n$_\n" unless length($token) eq length($line);
		}
		# Process statements
		elsif (exists $instr0{$token}) {
			# Instruction with no operands.  Easy case.
			die "Line: $linenum: Indirect not allowed" unless ($indirect eq 0);
			die "Line: $linenum: No-Load not allowed" unless ($noload eq 0);
			$instr = $instr0{$token};
			die "Line $linenum: No target allowed.\n$_\n" unless length($token) eq length($line);
		} elsif (exists $instrac34{$token}) {
			# Instruction that specifies an AC in bits 3&4
			die "Line: $linenum: Indirect not allowed" unless ($indirect eq 0);
			die "Line: $linenum: No-Load not allowed" unless ($noload eq 0);
			$instr = $instrac34{$token};
			$token = substr($line,length($token)+1);
			die "Line $linenum: Invalid AC: $token" unless (length($token) eq 1);
			die "Line $linenum: Invalid AC: $token" unless $token =~ /^[0-3]$/;
			$instr |= ($token << 11);
		} elsif (exists $instr2ac{$token}) {
			# Instruction that specifies 2 ACs
			die "Line: $linenum: Indirect not allowed" unless ($indirect eq 0);
			die "Line: $linenum: No-Load not allowed" unless ($noload eq 0);
			if ($token eq "OR") {
				$isor = 1;
			} else {
				$isor = 0;
			}
			$instr = $instr2ac{$token};
			$token = substr($line,length($token)+1);
			$token =~ s/ //g;
			die "Line $linenum: Invalid src/dst register" unless (exists $regsd{$token});
			$regs = $regsd{$token};
			if ($isor eq 0) {
				$instr |= $regs;
			} else {
				# Swap location of src & dst registers
				# Why was Fairchild inconsistent in placement???
				$instr |= (($regs << 5) & 014000) | (($regs >> 5) & 000300);
			}
		} elsif (exists $instrdev{$token}) {
			# Instructions that specify a device
			die "Line: $linenum: Indirect not allowed" unless ($indirect eq 0);
			die "Line: $linenum: No-Load not allowed" unless ($noload eq 0);
			$instr = $instrdev{$token};
			die "Line $linenum: No device specified" if (length($token) eq length($line));
			$token = substr($line,length($token)+1);
			if (exists $devs{$token}) {
				$instr |= $devs{$token};
			} else {
#				die "Line $linenum: Invalid device: $token" unless (length($token) le 2);
				die "Line $linenum: Invalid device: $token" unless $token =~ /^[0-7]{1,2}$/;
				$instr |= oct $token;
			}
		} elsif (exists $instrio{$token}) {
			# Instructions that specify AC & device
			die "Line: $linenum: Indirect not allowed" unless ($indirect eq 0);
			die "Line: $linenum: No-Load not allowed" unless ($noload eq 0);
			$instr = $instrio{$token};
			# Get AC
			$line = substr($line,length($token)+1);
			$zzz = index($line, ',');
			if ($zzz > 0) {
				$token = trim(substr($line,0,$zzz));
			}
			die "Line $linenum: Invalid AC" unless length($token) eq 1;
			die "Line $linenum: Invalid AC: $token" unless $token =~ /^[0-3]$/;
			$instr |= ($token << 11);
			$token = substr($line,$zzz+1);
			$token = trim($token);
			if (exists $devs{$token}) {
				$instr |= $devs{$token};
			} else {
#				die "Line $linenum: Invalid device: $token" unless (length($token) le 2);
				die "Line $linenum: Invalid device: $token" unless $token =~ /^[0-7]{1,2}$/;
				$instr |= oct $token;
			}
		} elsif (exists $instrtgt{$token}) {
			# Instructions that reference a mem loc.
			die "Line: $linenum: No-Load not allowed" unless ($noload eq 0);
			$instr = $instrtgt{$token} | $indirect;
			$token = substr($line,length($token)+1);
			# Target is either label, relative (./.+N/.-N) or absolute octal value
			#   OR indexed jump relative to AC2 or 3
			if ($token =~ /^\./) {
				$token =~ s/ //g;
				if ($token eq '.') {
					$instr |= 0400;
				} elsif ($token =~ /^\.\+/) {
					$token = substr($token,2);
					if ($token =~ /^\d+\.$/) {
						$token =~ s/\.//;
						$offset = $token;
					} else {
						die "Line $linenum: Invalid octal value: $token" unless $token =~ /^[0-7]+$/;
						$offset = oct $token;
					}
					die "Line $linenum: offset out of range" if $offset > 0177;
					$instr |= 0400 | $offset;
				} elsif ($token =~ /^\.\-/) {
					$token = substr($token,2);
					if ($token =~ /^\d+\.$/) {
						$token =~ s/\.//;
						$offset = $token;
					} else {
						die "Line $linenum: Invalid octal value: $token" unless $token =~ /^[0-7]+$/;
						$offset = oct $token;
					}
					die "Line $linenum: offset out of range" if $offset > 0200;
					$instr |= 0400 | ((-$offset) & 0377);
				} else {
					die "Line $linenum: Invalid relative address: $token";
				}
			} elsif (exists $sym_tab{$token}) {
				# A label!
				$offset = $sym_tab{$token};
				if ($offset > 0377) {
					$offset -= $addr;
					die "Line $linenum: target out of range" if $offset > 0177;
					die "Line $linenum: target out of range" if $offset < -0200;
					$offset &= 0377;
					$instr |= 0400;
				}
				$instr |= $offset;
			} elsif ($token =~/,/) {
				$offset = trim(substr($token,0,index($token,',')));
				die "Line $linenum: invalid offset: $token" unless $offset =~ /^[\+\-]?[0-7]+$/;
				if ($offset =~/^\+/) {
					$offset = substr($offset,1);
				}
				if ($offset =~/^\-/) {
					$offset = substr($offset,1);
					$offset = -oct($offset);
				} else {
					$offset = oct($offset);
				}
				die "Line $linenum: invalid offset" if $offset > 0177;
				die "Line $linenum: invalid offset" if $offset < -0200;
				$token = trim(substr($token,index($token,',') + 1));
				die "Line $linenum: invalid index register: $token" unless $token =~/^[23]$/;
				$instr |= ((oct($token) << 8) | ($offset & 0377));
			} else {
				# Octal or invalid
				die "Line $linenum: Invalid target address: $token" unless $token =~ /^[0-7]+$/;
				$offset = oct($token);
				if ($offset > 0377) {
					$offset -= $addr;
					die "Line $linenum: target out of range" if $offset > 0177;
					die "Line $linenum: target out of range" if $offset < -0200;
					$offset &= 0377;
					$instr |= 0400;
				}
				$instr |= $offset;
			}
		} elsif (exists $instractgt{$token}) {
			# Instructions that reference a mem loc & an AC
			die "Line: $linenum: No-Load not allowed" unless ($noload eq 0);
			$instr = $instractgt{$token} | $indirect;
			# Get AC
			$line = substr($line,length($token)+1);
			$zzz = index($line, ',');
			if ($zzz > 0) {
				$token = trim(substr($line,0,$zzz));
			}
			die "Line $linenum: Invalid AC" unless length($token) eq 1;
			die "Line $linenum: Invalid AC: $token" unless $token =~ /^[0-3]$/;
			$instr |= ($token << 11);
			$token = substr($line,$zzz+1);
			$token = trim($token);

			# Target is either label, relative (./.+N/.-N) or absolute octal value
			if ($token =~ /^\./) {
				$token =~ s/ //g;
				if ($token eq '.') {
					$instr |= 0400;
				} elsif ($token =~ /^\.\+/) {
					$token = substr($token,2);
					if ($token =~ /^\d+\.$/) {
						$token =~ s/\.//;
						$offset = $token;
					} else {
						die "Line $linenum: Invalid octal value: $token" unless $token =~ /^[0-7]+$/;
						$offset = oct $token;
					}
					die "Line $linenum: offset out of range" if $offset > 0177;
					$instr |= 0400 | $offset;
				} elsif ($token =~ /^\.\-/) {
					$token = substr($token,2);
					if ($token =~ /^\d+\.$/) {
						$token =~ s/\.//;
						$offset = $token;
					} else {
						die "Line $linenum: Invalid octal value: $token" unless $token =~ /^[0-7]+$/;
						$offset = oct $token;
					}
					die "Line $linenum: offset out of range" if $offset > 0200;
					$instr |= 0400 | ((-$offset) & 0377);
				} else {
					die "Line $linenum: Invalid relative address: $token";
				}
			} elsif (exists $sym_tab{$token}) {
				# A label!
				$offset = $sym_tab{$token};
				if ($offset > 0377) {
					$offset -= $addr;
					die "Line $linenum: target out of range" if $offset > 0177;
					die "Line $linenum: target out of range" if $offset < -0200;
					$offset &= 0377;
					$instr |= 0400;
				}
				$instr |= $offset;
			} elsif ($token =~/,/) {
				$offset = trim(substr($token,0,index($token,',')));
				die "Line $linenum: invalid offset: $token" unless $offset =~ /^[\+\-]?[0-7]+$/;
				if ($offset =~/^\+/) {
					$offset = substr($offset,1);
				}
				if ($offset =~/^\-/) {
					$offset = substr($offset,1);
					$offset = -oct($offset);
				} else {
					$offset = oct($offset);
				}
				die "Line $linenum: invalid offset" if $offset > 0177;
				die "Line $linenum: invalid offset" if $offset < -0200;
				$token = trim(substr($token,index($token,',') + 1));
				die "Line $linenum: invalid index register: $token" unless $token =~/^[23]$/;
				$instr |= ((oct($token) << 8) | ($offset & 0377));
			} else {
				# Octal or invalid
				die "Line $linenum: Invalid target address: $token" unless $token =~ /^[0-7]+$/;
				$offset = oct($token);
				if ($offset > 0377) {
					$offset -= $addr;
					die "Line $linenum: target out of range" if $offset > 0177;
					die "Line $linenum: target out of range" if $offset < -0200;
					$offset &= 0377;
					$instr |= 0400;
				}
				$instr |= $offset;
			}
		} elsif (exists $instralu{$token}) {
			# Instructions that reference 2 ACs + optional skip condition
			die "Line: $linenum: Indirect not allowed" unless ($indirect eq 0);
			$instr = $instralu{$token};
			$token = substr($line,length($token)+1);
			$token =~ s/ //g;
			die "Line: $linenum: Invalid registers/skip consition: $token" unless (exists $rskip{$token});
			$instr |= $rskip{$token};
			if ($noload eq 1) {
				$instr |= 000010;
			}
		} else {
			print "$line\n";
			die "Line $linenum: Unrecognized operation: $token";
		}

#		writehex($addr,$instr);
		writebin($addr,$instr);
		printf {$listfh} "%04d %06o %06o      %s",$linenum,$addr,$instr,$_;
		$addr++;
	}
	

	printf {$listfh} "\n\n\n\nSymbol Table:\n";
	foreach my $symbol (sort keys %main::sym_tab) {
		printf {$listfh} "%-16s %06o\n",$symbol,$main::sym_tab{$symbol};
	}

	close $in or die "$in: $!";
	close $listfh or die "$listfh: $!";
#	closehexfiles();
	closebinfile();

	my $end = Time::HiRes::gettimeofday();
	printf "Elapsed time: %.3f\n", $end-$start;
}


#######################################################################################
# Open the hex files for output
#######################################################################################
#sub openhexfiles {
#	my $hexfilename;
#	
#	$hexfilename = $progname . '-LO.HEX';
#	open($hexlofh, ">", $hexfilename) or die "Can't open hex file: $hexfilename: $!";
#	$hexfilename = $progname . '-HI.HEX';
#	open($hexhifh, ">", $hexfilename) or die "Can't open hex file: $hexfilename: $!";
#	return;
#}



#######################################################################################
# Close the hex files
#######################################################################################
#sub closehexfiles {
#	# Write end records
#	print {$hexlofh}  ":00000001FF\n";
#	print {$hexhifh}  ":00000001FF\n";
#	
#	close $hexlofh  or die "$hexlofh: $!";
#	close $hexhifh  or die "$hexhifh: $!";
#	return;
#}



#######################################################################################
# Write the current instruction to the hex files
#######################################################################################
#sub writehex {
#	my $record;
#	my $addrstr  = sprintf("%04X",shift);
#	my $instr    = shift;
#	my $lobyte   = $instr & 0xFF;
#	my $hibyte   = ($instr >> 8) & 0xFF;
#
#	# Reverse the bits in the bytes to match the way NOVA445 ROMs are wired
#	$hibyte = sprintf("%02X", ord pack('B8', unpack('b8', pack('C', $hibyte))));
#	$lobyte = sprintf("%02X", ord pack('B8', unpack('b8', pack('C', $lobyte))));
#
#	my $ckcom = hex("01") + hex(substr($addrstr,0,2)) + hex(substr($addrstr,2,2));
#
#	# Low byte record
#	my $checksum = - ($ckcom + hex($lobyte));
#	$record = sprintf(":01%s00%s%02X\n",$addrstr,$lobyte,$checksum & 0xFF);
#	print {$hexlofh} $record;
#
#	# High byte record
#	$checksum = - ($ckcom + hex($hibyte));
#	$record = sprintf(":01%s00%s%02X\n",$addrstr,$hibyte,$checksum & 0xFF);
#	print {$hexhifh} $record;
#	return;
#}


#######################################################################################
# Open the binary file for output
#######################################################################################
sub openbinfile {
	my $binfilename;

	$wcount = 0;
	$binfilename = $progname . '.AB';
	open($binfh, ">", $binfilename) or die "Can't open binary file: $binfilename: $!";
	return;
}



#######################################################################################
# Close the binary file
#######################################################################################
sub closebinfile {
	close $binfh  or die "$binfh: $!";
	return;
}



#######################################################################################
# Write the current instruction to the binary file buffer
#######################################################################################
sub writebin {
	shift;
	($datablk[$wcount++]) = @_;
	if ($wcount == 16) {
		flushbin();
	}
}


#######################################################################################
# Write the current binary file buffer to the file
#######################################################################################
sub flushbin {
	my $indx;
	my $checksum;

	if ($wcount == 0) {return};

	# Compute checksum
	$checksum = $wcount - $memaddr;
	for ($indx = 0; $indx < $wcount; $indx++) {
		$checksum -= $datablk[$indx];
	}

	# Now write data block
	writeword(-$wcount);
	writeword($memaddr);
	writeword($checksum);

	for ($indx = 0; $indx < $wcount; $indx++) {
		writeword($datablk[$indx]);
	}
	$memaddr += $wcount;
	$wcount = 0;
}


#######################################################################################
# Write a 16-bit word in binary to the file
#######################################################################################
sub writeword {
	my ($rec) = @_;
	$rec &= 0xFFFF;
	my $c = chr($rec & 0xFF);	# Low byte first
	syswrite $binfh,$c,1;
	$c = chr(($rec/256) & 0xFF);	# Now high byte
	syswrite $binfh,$c,1;

	return;
}


#######################################################################################
# Write an 8-bit byte in binary to the file
#######################################################################################
sub writebyte {
	my ($rec) = @_;
	my $c = chr($rec & 0xFF);	# Byte value
	syswrite $binfh,$c,1;

	return;
}


#######################################################################################
# Trim leading and trailing white space from a string
#######################################################################################
sub trim {
	my $s = shift;
	$s =~ s/^\s+|\s+$//g;
	return $s;
}


#######################################################################################
# Process angle brackets in a string constant
#######################################################################################
sub debracket {
	my $left;
	my $right;
	my $ch;

	my $s = shift;
	while ($s =~ /</) {
		$left = index($s,"<");
		$right = index($s, ">");
		die "Invalid use of angle brackets in string: $s" unless $right gt $left;
		# Extract character value
		$ch = substr($s,$left+1,$right-$left-1);
		# Verify it's octal
		die "Invalid <> value in string: $ch" unless $ch =~ /^[0-7]+$/;
		die "Invalid <> value in string: $ch" unless oct($ch) le 0377;
		# Make the substitution
		$s = substr($s,0,$left) . chr(oct($ch)) . substr($s,$right+1);
	}
	return $s;
}


#######################################################################################
# To Do:
#
# Single assembler for ROM & binary files?
